# coding: latin-1
# Copyright (c) 2014 Dirk Baechle.
# www: https://bitbucket.org/dirkbaechle/cmake2scons
# mail: dl9obn AT darc.de
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import os
import sys
import re
from pyparsing import *
import cmake2scons.cmakecmds as cmds


newline = Literal("\n")
quote = Literal('"')
left_paren = Suppress(Literal("("))
right_paren = Suppress(Literal(")"))

escape_semicolon = Literal("\\;")
escape_encoded = Regex(r'\t|\r|\n')
# TODO: re-add support for quoted parentheses
escape_identity = Regex(r'\#|\"|\ |\\|\$|\@|\^')
escape_sequence = Or([escape_identity, escape_encoded, escape_semicolon])

unquoted_element = Or([Regex(r'[^\s()#"\\]+'), escape_sequence])

# TODO: implement
# unquoted_legacy = ?TODO
# unquoted_argument = Or([OneOrMore(unquoted_element), unquoted_legacy])
unquoted_argument = OneOrMore(unquoted_element)
quoted_continuation = "\\" + newline

term_hash = Literal("#")

line_comment = Suppress(term_hash + restOfLine)
# TODO: escape bracket arguments
# line_comment ::=  '#' <any text not starting in a bracket_argument
#                       and not containing a newline>

quoted_element = Or([Regex("[^\\\"]+"), escape_sequence, quoted_continuation])
quoted_argument = quote + ZeroOrMore(quoted_element) + quote

# TODO: support for quoted brackets (new in CMake v3.0)
# bracket_argument ::=  bracket_open bracket_content bracket_close
# bracket_open     ::=  '[' '='{len} '['
# bracket_content  ::=  <any text not containing a bracket_close
#                       of the same {len} as the bracket_open>
# bracket_close    ::=  ']' '='{len} ']'
#
# bracket_comment ::=  '#' bracket_argument
#
# argument =  Or([bracket_argument, quoted_argument, unquoted_argument])

argument = Or([quoted_argument, unquoted_argument])

# line_ending = Optional(line_comment) + newline
arguments = Forward()
separated_arguments = Or([argument,
                          left_paren + arguments + right_paren])

arguments << (Optional(argument) + ZeroOrMore(separated_arguments))
identifier = Regex("[A-Za-z_][A-Za-z0-9_]*")
command_invocation = Group(identifier + left_paren + arguments + right_paren)
# TODO: support for bracket comments
# file_element = Or([command_invocation + line_ending,
#                   ZeroOrMore(bracket_comment) + line_ending])
file_element = Or([command_invocation + Optional(line_comment),
                   line_comment])

cmfile = ZeroOrMore(file_element)

# List of supported commands with their handlers.
supported = {'if$': cmds.handleIf,
             'elseif$': cmds.handleElseIf,
             'else$': cmds.handleElse,
             'endif$': cmds.handleEndif,
             'add_subdirectory$': cmds.handleAddSubdirectory,
             'add_([^_]+_)?tool_subdirectory$': cmds.handleAddToolSubdirectory,
             'add_([^_]+_)?executable$': cmds.handleAddExecutable,
             'add_([^_]+_)?tool$': cmds.handleAddTool,
             'add_([^_]+_)?library$': cmds.handleAddLibrary,
             'add_([^_]+_)?loadable_module$': cmds.handleAddLoadableModule,
             'include_directories$': cmds.handleIncludeDirectories,
             'link_directories$': cmds.handleLinkDirectories,
             'target_link_libraries$': cmds.handleTargetLinkLibraries,
             'set$': cmds.handleSet,
             'append_target_cflags$': cmds.handleAppendTargetCFlags,
             'set_target_link_flags$': cmds.handleSetTargetLinkFlags,
             'set_target_compile_flags$': cmds.handleSetTargetCompileFlags,
             'function$': cmds.handleFunction,
             'endfunction$': cmds.handleEndFunction,
             'option$': cmds.handleOption,
             'set_target_properties$': cmds.handleSetTargetProperties,
             'message$': cmds.handleMessage,
             'foreach$': cmds.handleForeach,
             'endforeach$': cmds.handleEndForeach
             }

# TODO
unknown = ['add_dependencies',
           'install',
           '([^_]+_)compile',
           '([^_]+_)link_shared',
           'add(_[^_])_external_project',
           'add(_[^_])_target',
           'add(_[^_])_executable',
           'add(_[^_])_utility',
           'add(_[^_])_example',
           'set_property',
           'get_property',
           'find_package',
           'parse_arguments',
           'export',
           'add_file_dependencies',
           'find_program']

# List of commands that will create a local Clone() of
# the global environment "env".
local = ['include_directories$',
         'link_directories$',
         'target_link_libraries$',
         'set$',
         'append_target_cflags$',
         'set_target_link_flags$',
         'set_target_compile_flags$',
         'set_target_properties$']

# List of commands that might rely on the possibly changed
# settings of the global environment, so an immediate
# Export('env') is required before the actual command...
# if there is one pending at all.
flushenv = ['add_subdirectory$',
            'add_([^_]+_)?tool_subdirectory$']

# Dict for managing global settings/targets/whatever...
environment = {}
# Set of commands that weren't handled during processing
unhandled_commands = set()
# Set of files that should be checked, because not all commands
# could be converted properly.
files_to_check = set()


def handleUnknownCommand(args, fout, env, opts):
    """ When encountering an unknown command, we simply paste
        the parsed command as Python comment to the output.
    """
    if fout:
        msg = " "*opts['indent']+"#"
        if args:
            msg += " %s" % args[0].lower()
        if len(args) > 1:
            msg += "(%s)" % ', '.join(args[1:])

        fout.write("%s\n" % msg)

    return env, opts


def needsLocalEnv(commands):
    """ Returns True if the list of parsed commands contains a
        keyword that requires a local build environment.
    """
    for c in commands:
        cmd = c[0].lower()
        # Search command in list of expressions that require
        # a local environment
        for k in local:
            m = re.match(k, cmd)
            if m:
                if cmd != "set":
                    return True
                else:
                    if len(c) > 1 and c[1].endswith("_LINK_COMPONENTS"):
                        return True

    return False


def convertCMake2SCons(fpath, spath, parse_only=False, toplevel=False):
    """ Converts the CMake build spec file fpath to the
        SConstruct/SConscript spath.
    """
    global cmfile
    global environment
    global unhandled_commands
    global files_to_check

    # Ensure that the global environment has a set member
    # to keep track of all the different options...
    if 'options' not in environment:
        environment['options'] = set()
    # Same for global variables/settings...
    if 'variables' not in environment:
        environment['variables'] = set()
    # ...and user-defined functions...
    if 'functions' not in environment:
        environment['functions'] = set()

    commands = cmfile.parseFile(fpath)

    if not commands:
        print "Error: couldn't parse file '%s' (empty?)" % fpath

    fout = None
    if not parse_only:
        fout = open(spath, "w")
        if toplevel:
            fout.write("import sys\n")
            fout.write("import os\n")
            fout.write("import re\n\n")
            fout.write("env = Environment()\n")
            fout.write("env['WIN32'] = (sys.platform.startswith('win'))\n")
            fout.write("env['UNIX'] = (sys.platform.startswith('linux')" +
                       " or sys.platform.startswith('darwin'))\n")
            fout.write("env['MSVC'] = (env['CC'] == 'cl')\n")
            fout.write("env['MINGW'] = False\n")
            fout.write("env['APPLE'] = False\n")
            fout.write("env['XCODE'] = False\n")
            fout.write("env['CYGWIN'] = False\n")
            fout.write("localEnv = env\n\n")
            environment['envexport'] = True
        else:
            fout.write("import sys\n")
            fout.write("import os\n")
            fout.write("import re\n\n")
            fout.write("Import('env')\n\n")
            if needsLocalEnv(commands):
                fout.write("localEnv = env.Clone()\n\n")
            else:
                fout.write("localEnv = env\n\n")

    # Collect commands
    options = {'indent': 0,
               'project': None,
               'file': fpath}
    for c in commands:
        cmd = c[0].lower()
        if environment.get('envexport', False):
            # Search command in list of expressions that require
            # the immediate re-export of the changed global environment
            for r in flushenv:
                m = re.match(r, cmd)
                if m:
                    # Export the environment and reset the flag
                    fout.write(" "*(options['indent']) +
                               "Export('env')\n")
                    environment['envexport'] = False
                    break
        # Search command in list of supported expressions
        found = False
        for k, h in supported.iteritems():
            m = re.match(k, cmd)
            if m:
                found = True
                g = m.groups()
                if len(g) > 1:
                    options['project'] = g[1][:-1]
                else:
                    options['project'] = None
                environment, options = h(c[1:], fout, environment, options)
                break
        if not found:
            # Could be a user-defined function
            if cmd in environment['functions']:
                environment, options = cmds.handleFunctionInvocation(c, fout,
                                                                     environment,
                                                                     options)
            else:
                unhandled_commands.add(cmd)
                files_to_check.add(spath)
                environment, options = handleUnknownCommand(c, fout,
                                                            environment,
                                                            options)

    if not parse_only:
        environment, options = cmds.handleEndOfFile(toplevel, fout,
                                                    environment, options)

        fout.close()


def convertAll(wdir="."):
    """ Converts a complete CMake project to SCons. """
    if wdir != ".":
        savedir = os.path.abspath(os.getcwd())
        os.chdir(wdir)

    for path, dirs, files in os.walk("."):
        for f in files:
            if f == "CMakeLists.txt":
                fpath = os.path.join(path, f)
                scfile = "SConscript"
                if not path or path == ".":
                    scfile = "SConstruct"
                    spath = os.path.join(path, scfile)
                    convertCMake2SCons(fpath, spath, False, True)
                else:
                    spath = os.path.join(path, scfile)
                    convertCMake2SCons(fpath, spath, False, False)

    if wdir != ".":
        os.chdir(savedir)


def main():
    wdir = "."
    if len(sys.argv) > 1:
        wdir = sys.argv[1]
    convertAll(wdir)

if __name__ == "__main__":
    main()
