# coding: latin-1
# Copyright (c) 2014 Dirk Baechle.
# www: https://bitbucket.org/dirkbaechle/cmake2scons
# mail: dl9obn AT darc.de
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

# Handlers for the different CMake commands


def noArgs(handler, opts):
    msg = "Warning: %s called without any arguments!" % handler
    msg += " (%s)" % opts.get('file', 'unknown')
    print msg


def tooFewArgs(handler, opts):
    msg = "Warning: %s called with too few arguments!" % handler
    msg += " (%s)" % opts.get('file', 'unknown')
    print msg


def groupArgs(args, lower=True):
    """ Return a list of single property tokens.

        Double quoted arguments are concatenated and treated
        as a single word.
    """
    vlist = []
    pq = 0
    word = []
    for a in args:
        # Hit a double quote
        if a == '"':
            if pq:
                # It's the end of the quoted argument
                pq = 0
                vlist.append(' '.join(word))
                word = []
            else:
                # Starting a new quoted argument
                pq = 1
        else:
            if pq:
                # Add this token to the current quoted argument
                word.append(a)
            else:
                # Add the token directly
                if lower:
                    a = a.lower()
                vlist.append(a)
    return vlist


def ifReplace(x, env):
    if x == "AND":
        return "and"
    elif x == "OR":
        return "or"
    elif x == "NOT":
        return "not"
    elif x == "LESS":
        return "<"
    elif x == "GREATER":
        return ">"
    elif x in ["WIN32", "UNIX", "MSVC",
               "MINGW", "APPLE", "XCODE",
               "CYGWIN"]:
        return "env['%s']" % x
    elif x in env['options']:
        return "env.GetOption('%s')" % x
    elif x in env['variables']:
        return "env['%s']" % x
    elif "${" in x:
        return "env.subst('%s')" % x
    return x


def ifQuote(arg):
    if arg.startswith('env['):
        return arg
    elif arg.startswith('env.subst('):
        return arg

    return "'%s'" % arg


def ifWrapFunctions(args):
    if len(args) > 1:
        wargs = []
        idx = 0
        while idx < len(args):
            a = args[idx].lower()
            if a == "matches":
                marg = wargs.pop()
                idx += 1
                if idx < len(args):
                    wargs.append("re.match(%s,%s)" % (ifQuote(args[idx]),
                                                      ifQuote(marg)))
            elif a == "strequal":
                marg = wargs.pop()
                idx += 1
                if idx < len(args):
                    wargs.append("%s == %s" % (ifQuote(marg),
                                               ifQuote(args[idx])))
            elif a == "exists":
                idx += 1
                if idx < len(args):
                    wargs.append("os.path.exists(%s)" % ifQuote(args[idx]))
            elif a == "defined":
                idx += 1
                if idx < len(args):
                    wargs.append("%s in env" % ifQuote(args[idx]))
            else:
                wargs.append(args[idx])

            idx += 1

        return wargs

    return args


def funcReplace(x, env):
    if x in env['options']:
        return "env.GetOption('%s')" % x
    elif x in env['variables']:
        return "env['%s']" % x
    return x


def boolReplace(x):
    if x in ["ON", "1"]:
        return "True"
    elif x in ["OFF", "0"]:
        return "False"
    return "'%s'" % x


def handleIf(args, fout, env, opts):
    if fout:
        if args:
            # filter arguments
            fargs = [ifReplace(x, env) for x in groupArgs(args, False)]
            fargs = ifWrapFunctions(fargs)
            fout.write(" "*(opts['indent'])+"if (%s):\n" % ' '.join(fargs))
        else:
            noArgs("handleIf")
            fout.write(" "*(opts['indent'])+"if 0:\n")

    opts['indent'] = opts['indent'] + 4
    return env, opts


def handleElse(args, fout, env, opts):
    if fout:
        fout.write(" "*(opts['indent']-4)+"else:\n")

    return env, opts


def handleElseIf(args, fout, env, opts):
    if fout:
        if args:
            # filter arguments
            fargs = [ifReplace(x, env) for x in args]
            fout.write(" "*(opts['indent']-4)+"elif (%s):\n" % ' '.join(fargs))
        else:
            noArgs("handleElseIf")
            fout.write(" "*(opts['indent']-4)+"elif 0:\n")

    return env, opts


def handleEndif(args, fout, env, opts):
    if fout:
        fout.write("\n")

    opts['indent'] = opts['indent'] - 4
    return env, opts


def handleAddSubdirectory(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.SConscript('%s/SConscript')\n" % args[0])
        else:
            noArgs("handleAddSubdirectory", opts)

    return env, opts


def handleAddToolSubdirectory(args, fout, env, opts):
    return handleAddSubdirectory(args, fout, env, opts)


def handleAddExecutable(args, fout, env, opts):
    if fout:
        if len(args) > 1:
            if len(args) > 2:
                fout.write(" "*(opts['indent']) +
                           'localEnv.Program("%s", Split("""%s"""))\n' %
                           (args[0], ' '.join(args[1:])))
            else:
                fout.write(" "*(opts['indent']) +
                           'localEnv.Program("%s", "%s")\n' %
                           (args[0], args[1]))
        else:
            noArgs("handleAddExecutable", opts)

    return env, opts


def handleAddTool(args, fout, env, opts):
    return handleAddExecutable(args, fout, env, opts)


def handleIncludeDirectories(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.Append(CPPPATH=['%s'])\n" % "', '".join(args))
        else:
            noArgs("handleIncludeDirectories", opts)

    return env, opts


def handleLinkDirectories(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.Append(LIBPATH=['%s'])\n" % "', '".join(args))
        else:
            noArgs("handleLinkDirectories", opts)

    return env, opts


def handleSet(args, fout, env, opts):
    if fout:
        if args:
            if len(args) > 1:
                if args[0].endswith("_LINK_COMPONENTS"):
                    fout.write(" "*(opts['indent']) +
                               "localEnv.Append(LIBS=['%s'])\n" %
                               "', '".join(args[1:]))
                else:
                    # Set a global option
                    fout.write(" "*(opts['indent']) +
                               "env['%s'] = %s\n" %
                               (args[0], boolReplace(args[1].strip('"'))))
                    env['envexport'] = True
            else:
                # Set a global option empty
                fout.write(" "*(opts['indent'])+"env['%s'] = ''\n" % args[0])
                env['envexport'] = True
        else:
            noArgs("handleSet", opts)

    return env, opts


def handleTargetLinkLibraries(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.Append(LIBS=['%s'])\n" % "', '".join(args))
        else:
            noArgs("handleTargetLinkLibraries", opts)

    return env, opts


def handleAddLibrary(args, fout, env, opts):
    if fout:
        if len(args) > 1:
            if len(args) > 2:
                fout.write(" "*(opts['indent']) +
                           'localEnv.Library("%s", Split("""%s"""))\n' %
                           (args[0], ' '.join(args[1:])))
            else:
                fout.write(" "*(opts['indent']) +
                           'localEnv.Library("%s", "%s")\n' %
                           (args[0], args[1]))
        else:
            noArgs("handleAddLibrary", opts)

    return env, opts


def handleAddLoadableModule(args, fout, env, opts):
    if fout:
        if len(args) > 1:
            if len(args) > 2:
                fout.write(" "*(opts['indent']) +
                           'localEnv.SharedLibrary("%s", Split("""%s"""))\n' %
                           (args[0], ' '.join(args[1:])))
            else:
                fout.write(" "*(opts['indent']) +
                           'localEnv.SharedLibrary("%s", "%s")\n' %
                           (args[0], args[1]))
        else:
            noArgs("handleAddLoadableModule", opts)

    return env, opts


def handleAppendTargetCFlags(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.Append(CXXFLAGS=['%s'])\n" %
                       "', '".join(args))
        else:
            noArgs("handleAppendTargetCFlags", opts)

    return env, opts


def handleSetTargetLinkFlags(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.Replace(LDFLAGS=['%s'])\n" %
                       "', '".join(args))
        else:
            noArgs("handleSetTargetLinkFlags", opts)

    return env, opts


def handleSetTargetCompileFlags(args, fout, env, opts):
    if fout:
        if args:
            fout.write(" "*(opts['indent']) +
                       "localEnv.Replace(CXXFLAGS=['%s'])\n" %
                       "', '".join(args))
        else:
            noArgs("handleSetTargetCompileFlags", opts)

    return env, opts


def handleFunction(args, fout, env, opts):
    if fout:
        if args:
            # filter arguments
            fargs = [funcReplace(x, env) for x in args[1:]]
            funcname = args[0].lower()
            fout.write(" "*(opts['indent']) +
                       "def %s(%s):\n" % (funcname, ', '.join(fargs)))
            fout.write(" "*(opts['indent']) +
                       "env.AddMethod('%s', %s)\n" % (funcname, funcname))
            env['functions'].add(args[0].lower())
            env['envexport'] = True
        else:
            noArgs("handleFunction", opts)

    opts['indent'] = opts['indent'] + 4
    return env, opts


def handleEndFunction(args, fout, env, opts):
    if fout:
        fout.write("\n")

    opts['indent'] = opts['indent'] - 4
    return env, opts


def handleFunctionInvocation(args, fout, env, opts):
    if fout:
        if args:
            # filter arguments
            fargs = [funcReplace(x, env) for x in args[1:]]
            fout.write(" "*(opts['indent']) +
                       "env.%s(%s)\n" % (args[0].lower(), ', '.join(fargs)))
        else:
            noArgs("handleFunctionInvocation", opts)

    return env, opts


def getOptionHelp(args, env, opts):
    help = ' '.join(args[1:-2])
    help = help.strip('"')
    help = help.strip()
    help = help.replace("'", "\\'")

    return help


def handleOption(args, fout, env, opts):
    if fout:
        if args:
            if len(args) > 2:
                value = args[-1].lower()
                if value in ['on', 'off']:
                    # Set a boolean option
                    if value == 'off':
                        act = 'store_false'
                        default = 'True'
                    else:
                        act = 'store_true'
                        default = 'False'
                    help = getOptionHelp(args, env, opts)
                    fout.write(" "*(opts['indent']) +
                               "AddOption('--%s', dest='%s', action='%s', help='%s', default=%s)\n" %
                               (args[0], args[0], act, help, default))
                    env['options'].add(args[0])
                    env['envexport'] = True
                else:
                    # Set string option
                    help = getOptionHelp(args, env, opts)
                    fout.write(" "*(opts['indent']) +
                               "AddOption('--%s', dest='%s', type='string', action='store', help='%s', default='%s')\n" %
                               (args[0], args[0], help, args[-1]))
                    env['options'].add(args[0])
                    env['envexport'] = True
            else:
                # Set a global option empty
                fout.write(" "*(opts['indent'])+"env['%s'] = ''\n" % args[0])
                env['variables'].add(args[0])
                env['envexport'] = True
        else:
            noArgs("handleSet", opts)

    return env, opts


def handleEndOfFile(toplevel, fout, env, opts):
    if not toplevel and env.get('envexport', False):
        fout.write("\nExport('env')\n\n")
        env['envexport'] = False

    return env, opts


def groupTargetProperties(args):
    props = {}
    if args[0].lower() == 'properties':
        # Create a list of single property tokens,
        # where double quoted arguments are treated
        # as a single word...
        vlist = groupArgs(args)
        # Now pick out the known/supported target options
        for idx, v in enumerate(vlist):
            if idx < (len(vlist)-1):
                if v == "compile_flags":
                    props['compile_flags'] = vlist[idx+1].split()
                elif v == "soversion":
                    props['soversion'] = [vlist[idx+1]]

    return props


def handleSetTargetProperties(args, fout, env, opts):
    if fout:
        if args:
            if len(args) > 2:
                # Group arguments into the single properties
                props = groupTargetProperties(args[1:])
                for p, v in props.iteritems():
                    if p == 'compile_flags':
                        fout.write(" "*(opts['indent']) +
                                   "localEnv.Append(CXXFLAGS=['%s'])\n" %
                                   "', '".join(v))
                    elif p == 'soversion':
                        fout.write(" "*(opts['indent']) +
                                   "localEnv['SHLIBVERSION']='%s')\n" %
                                   v[0])
            else:
                tooFewArgs("handleSetTargetProperties", opts)
        else:
            noArgs("handleSetTargetProperties", opts)

    return env, opts


def handleMessage(args, fout, env, opts):
    if fout:
        if args:
            # filter arguments
            msg = ' '.join(groupArgs(args[1:], False))
            fout.write(" "*(opts['indent']) +
                       "print '%s: %s'\n" % (args[0],
                                             msg.replace("'", "\\'")))
        else:
            noArgs("handleMessage", opts)

    return env, opts


def handleForeach(args, fout, env, opts):
    if fout:
        if args:
            # filter arguments
            fargs = [ifReplace(x, env) for x in args]
            fout.write(" "*(opts['indent'])+"for %s in %s:\n" % (fargs[0],
                                                                 fargs[1]))
        else:
            noArgs("handleForeach")
            fout.write(" "*(opts['indent'])+"if 0:\n")

    opts['indent'] = opts['indent'] + 4
    return env, opts


def handleEndForeach(args, fout, env, opts):
    if fout:
        fout.write("\n")

    opts['indent'] = opts['indent'] - 4
    return env, opts
